import pandas as pd


class ColumnExtractor(object):
    """Simple Class for extracting columns from a database.

    To be used in pipeline to apply the feature selection or elimination results
    """

    def __init__(self, cols):
        """Initialize the class with colums to keep."""
        self.cols = cols

    def transform(self, X):
        """Apply to a pd.DataFrame."""
        return X[self.cols]

    def fit(self, X, y=None):
        """No Fitter."""
        return self


class ColumnRenamer(object):
    """Simple Class for rename columns from a database.

    To be used in pipeline after transformers to allow further steps.
    """

    def __init__(self, col_names):
        """Initialize the class with columns name."""
        self.col_names = col_names

    def transform(self, X):
        """Apply to a np.array or pd.DataFrame."""
        X = pd.DataFrame(X, columns=self.col_names)
        return X

    def fit(self, X, y=None):
        """No Fitter."""
        return self



# Univariate filters
import pandas as pd
import numpy as np
from tqdm import tqdm
import sys
from cefeste.utils import get_categorical_features, time_step


@time_step
def find_constant_features(df, feat_to_check=None):
    """Retrieve the list of features that are constant.

    Args:
        df (pd.DataFrame): DataFrame to be checked
        feat_to_check (list): list of features to be checked
    Returns:
        list: list of constant features
    """
    if feat_to_check is None:
        feat_to_check = []

    if len(feat_to_check) == 0:
        feat_to_check = df.columns

    feat_to_remove = df[feat_to_check].loc[:, lambda db: (db.nunique() == 1)].columns.tolist()

    return feat_to_remove


@time_step
def find_low_nvalues_features(df, min_unique_val=2, feat_to_check=None):
    """Retrieve the list of features that have a number of distinct values lower than the minimum.

    If min_unique_val = 2 the function gets the same result as find_constant_features.

    Args:
        df (pd.DataFrame): DataFrame to be checked
        min_unique_val (int): minimum number of distinct values
        feat_to_check (list): list of features to be checked
    Returns:
        list: list of constant features
    """
    if feat_to_check is None:
        feat_to_check = []

    if len(feat_to_check) == 0:
        feat_to_check = df.columns
    feat_to_remove = df[feat_to_check].loc[:, lambda db: (db.nunique() < min_unique_val)].columns.tolist()

    return feat_to_remove


@time_step
def find_missing_features(df, max_pct_missing=0.9, feat_to_check=None):
    """Retrieve the list of features that have missings percentage higher than threshold.

    Args:
        df (pd.DataFrame): DataFrame to be checked
        max_pct_missing (float): threshold of missings percentage acceptable
        feat_to_check (list): list of features to be checked
    Returns:
        list: list of constant features
    """
    if feat_to_check is None:
        feat_to_check = []

    if len(feat_to_check) == 0:
        feat_to_check = df.columns

    feat_to_remove = (
        df[feat_to_check].loc[:, lambda db: (db.isna().sum(axis=0) / db.shape[0] > max_pct_missing)].columns.tolist()
    )

    return feat_to_remove


def find_high_topcat_features(df, max_pct_mfv=0.95, feat_to_check=None):
    """Retrieve the list of features that are equals to a unique value for too many observations.

    Args:
        df (pd.DataFrame): DataFrame to be checked
        max_pct_mfv (float): Maximum percentage of the Most Frequent Value acceptable
        feat_to_check (list): list of features to be checked
    Returns:
        list: list of constant features
    """
    if feat_to_check is None:
        feat_to_check = []

    if len(feat_to_check) == 0:
        feat_to_check = df.columns
    feat_to_check_pb = tqdm(feat_to_check, file=sys.stdout, desc="Performing find_high_topcat_features", ncols=100)
    feat_to_remove = []
    for ix, col in enumerate(feat_to_check_pb):
        if df.loc[df[col] == df[col].mode()[0]].shape[0] / df.shape[0] > max_pct_mfv:
            feat_to_remove.append(col)
        if ix == len(feat_to_check) - 1:
            feat_to_check_pb.set_description("Completed find_high_topcat_features", refresh=True)

    return feat_to_remove


def find_unstable_psi_features(
    base_df, compare_df, max_psi=0.2, psi_nbins=20, psi_bin_min_pct=0.02, feat_to_check=None
):
    """Retrieve the list of features that are unstable between sets (usually train and test).

    Args:
        base_df (pd.DataFrame): reference dataset, usually the train set.
        compare_df (pd.DataFrame): comparison dataset, usually the test set.
        max_psi (float): maximum psi acceptable
        psi_nbins (int): number of bins into which the features will be bucketed (maximum) to compute psi
        psi_bin_min_pct (float): minimum percentage of observations per bucket
        feat_to_check (list): list of features to be checked
    Returns:
        list: list of constant features
    """
    if feat_to_check is None:
        feat_to_check = []

    psi_results = retrieve_psis(
        base_df, compare_df, feat_to_check, psi_nbins=psi_nbins, psi_bin_min_pct=psi_bin_min_pct
    )

    feat_to_remove = [x for x, y in psi_results.items() if y > max_psi]

    return feat_to_remove


def retrieve_psis(base_df, compare_df, features, psi_nbins=20, psi_bin_min_pct=0.02):
    """Retrieve psi for all the features.

    Args:
        base_df (pd.DataFrame): reference dataset, usually the train set.
        compare_df (pd.DataFrame): comparison dataset, usually the test set.
        psi_nbins (int): number of bins into which the features will be bucketed (maximum) to compute psi
        psi_bin_min_pct (float): minimum percentage of observations per bucket
        features (list): list of features to be checked
    Returns:
        dict: dictionary reporting for each feature (key) the psi value (value)
    """
    if len(features) == 0:
        features = base_df.columns

    psis = dict()
    features_pb = tqdm(features, file=sys.stdout, desc="Performing find_unstable_psi_features", ncols=100, leave=True)
    for ix, col in enumerate(features_pb):
        if col in get_categorical_features(base_df):
            base_df = base_df.fillna("Missing")
            compare_df = compare_df.fillna("Missing")
            mapper = retrieve_bin_categorical(base_df, col, max_n_bins=psi_nbins)
            mapper = merge_categorical_bins(base_df, col, mapper, bin_min_pct=psi_bin_min_pct)
            base_bin = base_df[col].map(mapper)
            comp_bin = compare_df[col].map(mapper)
        else:
            base_df = base_df.fillna(-999999)
            compare_df = compare_df.fillna(-999999)
            cuts = retrieve_bin_numerical(base_df, col, max_n_bins=psi_nbins)
            cuts = merge_numerical_bins(base_df, col, cuts, bin_min_pct=psi_bin_min_pct)
            base_bin = pd.cut(base_df[col], cuts, right=True, duplicates="drop")
            comp_bin = pd.cut(compare_df[col].astype("float"), cuts, right=True, duplicates="drop")
        if ix == len(features) - 1:
            features_pb.set_description("Completed find_unstable_psi_features", refresh=True)
        psis[col] = psi_value(base_bin, comp_bin)

    return psis


def retrieve_bin_categorical(df, feat, max_n_bins=20):
    """Retrieve the bucket for a categorical feature.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        max_n_bins (int): number of bins into which the features will be bucketed (maximum) to compute psi
    Returns:
        dict: mapper to be used to bucket the feature feat
    """
    if df[feat].nunique() > max_n_bins:
        freqs = df[feat].value_counts(ascending=False)
        mapper_1_1 = {x: x for x in freqs.index[: max_n_bins - 1]}
        mapper_other = {x: "_other_" for x in freqs.index[max_n_bins - 1 :]}
        mapper = {**mapper_1_1, **mapper_other}
    else:
        mapper = {x: x for x in df[feat].unique()}

    return mapper


def merge_categorical_bins(df, feat, mapper, bin_min_pct=0.02):
    """Merge buckets that have low frequency.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        mapper (dict): initial mapper to be used to bucket the feature feat
        bin_min_pct (float): minimum percentage of observations per bucket

    Returns:
        dict: mapper to be used to bucket the feature feat
    """
    db = df[[feat]].assign(bin=df[feat].map(mapper))

    freqs = db["bin"].value_counts(ascending=True, normalize=True)

    if freqs.iloc[0] < bin_min_pct:
        update = {x: freqs.index[1] for x, y in mapper.items() if y == freqs.index[0]}
        mapper.update(update)
        mapper = merge_categorical_bins(df, feat, mapper, bin_min_pct)

    return mapper


def retrieve_bin_numerical(df, feat, max_n_bins=20):
    """Retrieve the bucket for a categorical feature.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        max_n_bins (int): number of bins into which the features will be bucketed (maximum) to compute psi
    Returns:
        list: the list of cuts to be used in pd.cut
    """
    if df[feat].nunique() > max_n_bins:
        db = df[[feat]].assign(bucket=pd.qcut(df[feat], q=max_n_bins, labels=False, duplicates="drop"))
        cuts = list(db.groupby("bucket")[feat].max())
    else:
        cuts = list(df[feat].unique())

    cuts.sort()
    cuts = [-np.Inf] + cuts
    cuts[-1] = np.Inf

    return cuts


def merge_numerical_bins(df, feat, cuts, bin_min_pct=0.02):
    """Merge buckets that have low frequency.

    Args:
        df (pd.DataFrame): DataFrame to be used
        feat (str): feature to be bucketed
        cuts (list): initial cuts to be used to bucket the feature feat
        bin_min_pct (float): minimum percentage of observations per bucket

    Returns:
        dict: mapper to be used to bucket the feature feat
    """
    db = df[[feat]].assign(bin=pd.cut(df[feat], bins=cuts, labels=False, right=True))
    freqs = db["bin"].value_counts(ascending=True, normalize=True)

    if freqs.iloc[0] < bin_min_pct:
        remove_cut = min(min(cuts[1:]), max(max(cuts[:-1]), freqs.index[0]))
        cuts.remove(remove_cut)
        cuts = merge_numerical_bins(db, feat, cuts, bin_min_pct=bin_min_pct)

    return cuts


def psi_value(ref, new):
    """Retrieve the psi.

    Args:
        ref (pd.Series): series of the features in the base/reference dataset
        new (pd.Series): series of the features in the new/comparison dataset
    Returns:
        float: the psi
    """
    ref_count = pd.DataFrame(ref.value_counts(normalize=True))
    ref_count.columns = ["ref"]
    new_count = pd.DataFrame(new.value_counts(normalize=True))
    new_count.columns = ["new"]

    merged = ref_count.merge(new_count, how="left", left_index=True, right_index=True).fillna(0)
    # merged.clip(lower=0.000001)

    merged["psi"] = (merged["new"] - merged["ref"]) * (np.log((merged["new"] / merged["ref"]) + 1e-9))
    return merged["psi"].sum()



from cefeste import FeatureAnalysis
from cefeste.utils import get_categorical_features
from cefeste.elimination.shap_rfe import Shap_RFE_full
import warnings

from sklearn.model_selection import RandomizedSearchCV, StratifiedKFold


class FeatureElimination(FeatureAnalysis):
    """Child class of FeatureAnalysis for Feature Elimination."""

    def __init__(
        self,
        # DB / Feature Parameters / Model
        db,
        target_col,
        model,
        grid,
        feat_to_check=None,
        sample_col=None,
        sample_train_value=None,
        algo_type="classification",
        # Hyperparameters Tuning / Cross Validation Fold
        cv_funct=RandomizedSearchCV,
        cv_scoring="auto",
        n_iter=20,
        manage_groups=False,
        groups=None,
        cv_type=StratifiedKFold(5, random_state=42, shuffle=True),
        manage_cats=False,
        # Reporting
        step_size=0.1,
        min_n_feat_step=5,
        final_n_feature=1,
        verbose=True,
        write_final=False,
        write_substep=False,
        dim_cat_threshold=10,
    ):
        """Feature Elimination Class.

        Args:
            db (pd.DataFrame): DataFrame to analyze.
            target_col (str): Name of the target column.
            model : classifier or regressor in sklearn API class.
            grid (dict): hyperparameters grid.
            feat_to_check (list, optional): Feature to analyze. If None all are used. Defaults to None.
            sample_col (str, optional): Name of the columns that indicates the samples (train, test). Defaults to None.
            sample_train_value (str, optional): Value of the sample column that indicates the train set. Defaults to None.
            algo_type (str, optional): "classification", "multiclass", "regression", describes the problem type.
                "classification" has to be used only for binary classification. Defaults to "classification".
            cv_funct: function or class for the Cross Validation. Defaults to RandomizedSearchCV
            cv_scoring: scoring argument of the cv_functs. Defaults to "auto" selects "roc_auc" fo classification, "r2" for regression 
                and "balanced accuracy" for multiclass.
            n_iter (int, optional): number of iteration, i.e. set of hyperparams tested in Cross Validation. Defaults to 20.
            manage_groups (bool, optional): determines if there is a feature whose groups have to be kept joined in CV. Defaults to False.
            groups (pd.Series, optional): feature whose groups have to be kept joined in CV. Defaults to None.
            cv_type : function or class for defining the CV sets. Defaults to StratifiedKFold(5, random_state=42, shuffle=True).
            manage_cats (bool, optional): determines if there are categorical features to be preprocessed. Defaults to False.
            step_size (int or float, optional): determines how many features to remove at each step.
                Fixed int or percentage of total features. Defaults to 0.1.
            min_n_feat_step (int, optional): min number of feature to remove at each step. Defaults to 5.
            final_n_feature (int, optional): number of features of the last model. Defaults to 1.
            verbose (bool, optional): If True print a log of information. Defaults to True.
            write_final (bool, optional): If True it saves in the current directory the final report that can be used for quick start. Defaults to False.
            write_substep (bool, optional): If True it saves in the current directory the report after SHAP_RFE that can be used for quick start. Defaults to False.
            dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
                Defaults to 10.

        """
        super().__init__(db, feat_to_check)
        self.feat_to_check = list(set(self.feat_to_check) - set([target_col, sample_col]))
        # Target check
        if (target_col in db.columns) | (target_col is None):
            self.target_col = target_col
        else:
            raise ValueError(f"{target_col} not in DataFrame")

        # Check the target var and set the algo type
        if algo_type not in ["auto", "regression", "classification", "multiclass", "unsupervised"]:
            raise ValueError(
                f"{algo_type} is not a valid algo_type. It should be one of the following:\n ['auto', 'regression', 'classification', 'multiclass', 'unsupervised']"
            )

        if target_col is not None:
            vals = db[target_col].nunique()
            if vals == 1:
                raise ValueError(f"The target column {target_col} selected is constant")
            elif algo_type != "auto":
                self.algo_type = algo_type
            elif vals == 2:
                self.algo_type = "classification"
            elif vals < 11:
                self.algo_type = "multiclass"
            else:
                self.algo_type = "regression"
        else:
            self.algo_type = "unsupervised"

        # Sample column check
        if (sample_col in db.columns) | (sample_col is None):
            self.sample_col = sample_col
        else:
            raise ValueError(f"{sample_col} not in DataFrame")
        # Save the list of samples in db
        if sample_col is not None:
            self._sample_list = db[sample_col].unique()
            self._n_sample = db[sample_col].nunique()
        else:
            self._sample_list = []
            self._n_sample = None
        # Sample Train value
        if (sample_train_value is not None) & (sample_train_value not in self._sample_list):
            raise ValueError(
                f"The value {sample_train_value} set for parameter sample_train_value is not in {sample_col}."
            )
        else:
            self.sample_train_value = sample_train_value

        # Parameters
        self.model = model
        self.grid = grid
        self.cv_funct = cv_funct
        self.cv_scoring = cv_scoring
        self.n_iter = n_iter
        self.manage_groups = manage_groups
        self.groups = groups
        self.cv_type = cv_type
        self.manage_cats = manage_cats
        self.step_size = step_size
        self.min_n_feat_step = min_n_feat_step
        self.final_n_feature = final_n_feature
        self.verbose = verbose
        self.write_final = write_final
        self.write_substep = write_substep
        self.dim_cat_threshold = dim_cat_threshold
        self.report = None
        self.final_feat = []
        self.selection_rule = None
        self.number_feat_rep = None
        self._filtered_out_features = []

    def make_report(self, **kwargs):
        """Perform the SHAP RFE report.

        Returns:
            pd.DataFrame: shap report
        """
        self.set_params(**kwargs)
        # for k, v in kwargs.items():
        #    self.__dict__[k] = v

        db = self.db
        db.reset_index(inplace=True, drop=True)

        if self.sample_col is not None:
            df, sample_series = self.db.drop(columns=self.sample_col), self.db[self.sample_col]
        else:
            df = db

        if self.target_col is not None:
            X, y = df.drop(columns=self.target_col), df[self.target_col]
        else:
            raise ValueError("shap rfe not performed since no target column defined")
        X = X[self.feat_to_check]

        if self.sample_train_value is not None:
            X_train = X.loc[sample_series == self.sample_train_value]
            y_train = y.loc[sample_series == self.sample_train_value]
        else:
            X_train = X
            y_train = y

        # Groups check
        if self.manage_groups:
            if self.groups is None:
                warnings.warn("no group defined")
                self.manage_groups = False
            elif not self.groups.index.equals(self.db.index):
                raise ValueError("Groups Series index do not match with DataFrame index in input!")
            else:
                self.groups = self.groups.reset_index(drop=True).iloc[X_train.index]
        else:
            self.groups = None

        # Categorical Features
        if self.manage_cats:
            self.categorical_features_list = get_categorical_features(X_train)
        else:
            self.categorical_features_list = []

        shap_report = Shap_RFE_full(
            X_train,
            y_train,
            model=self.model,
            grid=self.grid,
            cv_funct=self.cv_funct,
            cv_scoring=self.cv_scoring,
            n_iter=self.n_iter,
            manage_groups=self.manage_groups,
            groups=self.groups,
            cv_type=self.cv_type,
            algo_type=self.algo_type,
            step_size=self.step_size,
            min_n_feat_step=self.min_n_feat_step,
            final_n_feature=self.final_n_feature,
            verbose=self.verbose,
            write_final=self.write_final,
            write_substep=self.write_substep,
            manage_cats=self.manage_cats,
            categorical_features_list=self.categorical_features_list,
            dim_cat_threshold=self.dim_cat_threshold,
        )

        self.report = shap_report

        return self.report

    def shap_plot(self):
        """Plot the report."""
        if self.report is not None:
            self.report.plot(
                x="n_feat", y=["train_score", "valid_score"], xlim=(max(self.report.n_feat), min(self.report.n_feat))
            )
        else:
            raise ValueError("Missing report, run .make_report() first")

    def extract_features(self, selection_rule="decrease_perf", number_feat_rep=None, gap=0.1):
        """Extract survived features after SHAP RFE.

        Args:
            selection_rule (str, optional): "decrease_perf", "best_valid", "num_feat", describes the rule for which the features after shap are chosen. Defaults to 'decrease_perf'.
            number_feat_rep (int, optional): valid only for the selection_rule "num_feat", determines the number of features to be extracted according to the report. Defaults to None.
            gap(float, optional): valid only for the selection_rule "decrease_perf", identifies the maximum gap over which the decrease in performance is not acceptable.
        Returns:
            list: list of features extracted after SHAP RFE according to the selection_rule.
        """
        self.selection_rule = selection_rule
        self.number_feat_rep = number_feat_rep
        # Check the target var and set the algo type
        if selection_rule not in ["decrease_perf", "best_valid", "num_feat"]:
            raise ValueError(
                f"{selection_rule} is not a valid selection_rule. It should be one of the following:\n ['decrease_perf', 'best_valid', 'num_feat']"
            )

        if self.report is not None:
            if selection_rule == "decrease_perf":
                # Define n_feat: the first time the validation score decreases of more than 10%
                cutoff = (
                    self.report[["n_feat", "valid_score"]]
                    .sort_values("n_feat", ascending=False)
                    .assign(max_until=lambda x: x.valid_score.expanding().max())
                    .assign(valid_next=lambda x: x.valid_score.shift(periods=-1))
                    .assign(
                        gap=lambda x: x.apply(
                            lambda y: (y.max_until / y.valid_next) - 1
                            if y.valid_next > 0
                            else 1 - (y.max_until / y.valid_next),
                            axis=1,
                        )
                    )
                    .loc[lambda x: x.gap > gap]
                )
                if cutoff.shape[0] < 1:
                    cutoff = self.report.loc[self.report.n_feat.idxmin(), "n_feat"]
                else:
                    cutoff = cutoff.loc[lambda x: x.n_feat.idxmax(), "n_feat"]

                self.final_feat = list(self.report.loc[self.report["n_feat"] == cutoff, "feat_used"])[0]

            elif selection_rule == "best_valid":
                # Define n_feat: best validation score iteration
                n_min_feat = self.report.loc[
                    self.report["valid_score"] == self.report["valid_score"].max(), "n_feat"
                ].min()
                self.final_feat = list(self.report.loc[self.report["n_feat"] == n_min_feat, "feat_used"])[0]

            else:
                # Define n_feat: the user choose the features to use according to number of features to be used
                if number_feat_rep is not None:
                    try:
                        self.final_feat = list(self.report.loc[self.report["n_feat"] == number_feat_rep, "feat_used"])[
                            0
                        ]
                    except Exception:
                        raise ValueError(f"{number_feat_rep} number of features chosen uncorrect, look the report")
                else:
                    self.final_feat = list(
                        self.report.loc[self.report["n_feat"] == self.report["n_feat"].max(), "feat_used"]
                    )[0]
        else:
            raise ValueError("Missing report, run .make_report() first")

        self._filtered_out_features = list(set(self.feat_to_check) - set(self.final_feat))
        return self.final_feat

    def get_X_original(self):
        """Return the dataset with just the features to check."""
        return self.db[self.feat_to_check]

    def get_X_reduced(self):
        """Return the reduced dataset without the sample_col, target_col and the filtered features."""
        return self.db[self.final_feat]

    def get_db_filtered(self):
        """Return the reduced dataset without the filtered features."""
        return self.db.drop(columns=self._filtered_out_features)
